@@ -3,20 +3,20 @@ import re
 import datetime
 import sys
 from functools import wraps
-
+from decimal import Decimal, InvalidOperation
 
 try:
     from schema_builder import Schema, raises, message
     from error import (MultipleInvalid, CoerceInvalid, TrueInvalid, FalseInvalid, BooleanInvalid, Invalid, AnyInvalid,
                        AllInvalid, MatchInvalid, UrlInvalid, EmailInvalid, FileInvalid, DirInvalid, RangeInvalid,
-                       PathInvalid, ExactSequenceInvalid, LengthInvalid, DatetimeInvalid, InInvalid, TypeInvalid,
-                       NotInInvalid)
+                       PathInvalid, ExactSequenceInvalid, LengthInvalid, DatetimeInvalid, DateInvalid, InInvalid,
+                       TypeInvalid, NotInInvalid, ContainsInvalid)
 except ImportError:
     from .schema_builder import Schema, raises, message
     from .error import (MultipleInvalid, CoerceInvalid, TrueInvalid, FalseInvalid, BooleanInvalid, Invalid, AnyInvalid,
                         AllInvalid, MatchInvalid, UrlInvalid, EmailInvalid, FileInvalid, DirInvalid, RangeInvalid,
-                        PathInvalid, ExactSequenceInvalid, LengthInvalid, DatetimeInvalid, InInvalid, TypeInvalid,
-                        NotInInvalid)
+                        PathInvalid, ExactSequenceInvalid, LengthInvalid, DatetimeInvalid, DateInvalid, InInvalid,
+                        TypeInvalid, NotInInvalid, ContainsInvalid)
 
 
 if sys.version_info >= (3,):
@@ -416,8 +416,13 @@ def IsFile(v):
     True
     >>> with raises(FileInvalid, 'not a file'):
     ...   IsFile()("random_filename_goes_here.py")
+    >>> with raises(FileInvalid, 'Not a file'):
+    ...   IsFile()(None)
     """
-    return os.path.isfile(v)
+    if v:
+        return os.path.isfile(v)
+    else:
+        raise FileInvalid('Not a file')
 
 
 @message('not a directory', cls=DirInvalid)
@@ -427,8 +432,13 @@ def IsDir(v):
 
     >>> IsDir()('/')
     '/'
+    >>> with raises(DirInvalid, 'Not a directory'):
+    ...   IsDir()(None)
     """
-    return os.path.isdir(v)
+    if v:
+        return os.path.isdir(v)
+    else:
+        raise DirInvalid("Not a directory")
 
 
 @message('path does not exist', cls=PathInvalid)
@@ -440,8 +450,42 @@ def PathExists(v):
     True
     >>> with raises(Invalid, 'path does not exist'):
     ...   PathExists()("random_filename_goes_here.py")
+    >>> with raises(PathInvalid, 'Not a Path'):
+    ...   PathExists()(None)
     """
-    return os.path.exists(v)
+    if v:
+        return os.path.exists(v)
+    else:
+        raise PathInvalid("Not a Path")
+
+
+class Maybe(object):
+    """Validate that the object is of a given type or is None.
+
+    :raises Invalid: if the value is not of the type declared and is not None
+
+    >>> s = Schema(Maybe(int))
+    >>> s(10)
+    10
+    >>> with raises(Invalid):
+    ...  s("string")
+
+    """
+    def __init__(self, kind, msg=None):
+        if not isinstance(kind, type):
+            raise TypeError("kind has to be a type")
+
+        self.kind = kind
+        self.msg = msg
+
+    def __call__(self, v):
+        if v is not None and not isinstance(v, self.kind):
+            raise Invalid(self.msg or "%s must be None or of type %s" % (v, self.kind))
+
+        return v
+
+    def __repr__(self):
+        return 'Maybe(%s)' % str(self.kind)
 
 
 class Range(object):
@@ -475,19 +519,19 @@ class Range(object):
 
     def __call__(self, v):
         if self.min_included:
-            if self.min is not None and v < self.min:
+            if self.min is not None and not v >= self.min:
                 raise RangeInvalid(
                     self.msg or 'value must be at least %s' % self.min)
         else:
-            if self.min is not None and v <= self.min:
+            if self.min is not None and not v > self.min:
                 raise RangeInvalid(
                     self.msg or 'value must be higher than %s' % self.min)
         if self.max_included:
-            if self.max is not None and v > self.max:
+            if self.max is not None and not v <= self.max:
                 raise RangeInvalid(
                     self.msg or 'value must be at most %s' % self.max)
         else:
-            if self.max is not None and v >= self.max:
+            if self.max is not None and not v < self.max:
                 raise RangeInvalid(
                     self.msg or 'value must be lower than %s' % self.max)
         return v
@@ -504,6 +548,7 @@ class Clamp(object):
     """Clamp a value to a range.
 
     Either min or max may be omitted.
+
     >>> s = Schema(Clamp(min=0, max=1))
     >>> s(0.5)
     0.5
@@ -572,6 +617,29 @@ class Datetime(object):
         return 'Datetime(format=%s)' % self.format
 
 
+class Date(Datetime):
+    """Validate that the value matches the date format."""
+
+    DEFAULT_FORMAT = '%Y-%m-%d'
+    FORMAT_DESCRIPTION = 'yyyy-mm-dd'
+
+    def __call__(self, v):
+        try:
+            datetime.datetime.strptime(v, self.format)
+            if len(v) != len(self.FORMAT_DESCRIPTION):
+                raise DateInvalid(
+                    self.msg or 'value has invalid length'
+                                ' expected length %d (%s)' % (len(self.FORMAT_DESCRIPTION), self.FORMAT_DESCRIPTION))
+        except (TypeError, ValueError):
+            raise DateInvalid(
+                self.msg or 'value does not match'
+                            ' expected format %s' % self.format)
+        return v
+
+    def __repr__(self):
+        return 'Date(format=%s)' % self.format
+
+
 class In(object):
     """Validate that a value is in a collection."""
 
@@ -612,6 +680,33 @@ class NotIn(object):
         return 'NotIn(%s)' % (self.container,)
 
 
+class Contains(object):
+    """Validate that the given schema element is in the sequence being validated.
+
+    >>> s = Contains(1)
+    >>> s([3, 2, 1])
+    [3, 2, 1]
+    >>> with raises(ContainsInvalid, 'value is not allowed'):
+    ...   s([3, 2])
+    """
+
+    def __init__(self, item, msg=None):
+        self.item = item
+        self.msg = msg
+
+    def __call__(self, v):
+        try:
+            check = self.item not in v
+        except TypeError:
+            check = True
+        if check:
+            raise ContainsInvalid(self.msg or 'value is not allowed')
+        return v
+
+    def __repr__(self):
+        return 'Contains(%s)' % (self.item, )
+
+
 class ExactSequence(object):
     """Matches each element in a sequence against the corresponding element in
     the validators.
@@ -634,7 +729,7 @@ class ExactSequence(object):
         self._schemas = [Schema(val, **kwargs) for val in validators]
 
     def __call__(self, v):
-        if not isinstance(v, (list, tuple)):
+        if not isinstance(v, (list, tuple)) or len(v) != len(self._schemas):
             raise ExactSequenceInvalid(self.msg)
         try:
             v = type(v)(schema(x) for x, schema in zip(v, self._schemas))
@@ -691,3 +786,150 @@ class Unique(object):
 
     def __repr__(self):
         return 'Unique()'
+
+
+class Equal(object):
+    """Ensure that value matches target.
+
+    >>> s = Schema(Equal(1))
+    >>> s(1)
+    1
+    >>> with raises(Invalid):
+    ...    s(2)
+
+    Validators are not supported, match must be exact:
+
+    >>> s = Schema(Equal(str))
+    >>> with raises(Invalid):
+    ...     s('foo')
+    """
+
+    def __init__(self, target, msg=None):
+        self.target = target
+        self.msg = msg
+
+    def __call__(self, v):
+        if v != self.target:
+            raise Invalid(self.msg or 'Values are not equal: value:{} != target:{}'.format(v, self.target))
+        return v
+
+    def __repr__(self):
+        return 'Equal({})'.format(self.target)
+
+
+class Unordered(object):
+    """Ensures sequence contains values in unspecified order.
+
+    >>> s = Schema(Unordered([2, 1]))
+    >>> s([2, 1])
+    [2, 1]
+    >>> s([1, 2])
+    [1, 2]
+    >>> s = Schema(Unordered([str, int]))
+    >>> s(['foo', 1])
+    ['foo', 1]
+    >>> s([1, 'foo'])
+    [1, 'foo']
+    """
+
+    def __init__(self, validators, msg=None, **kwargs):
+        self.validators = validators
+        self.msg = msg
+        self._schemas = [Schema(val, **kwargs) for val in validators]
+
+    def __call__(self, v):
+        if not isinstance(v, (list, tuple)):
+            raise Invalid(self.msg or 'Value {} is not sequence!'.format(v))
+
+        if len(v) != len(self._schemas):
+            raise Invalid(self.msg or 'List lengths differ, value:{} != target:{}'.format(len(v), len(self._schemas)))
+
+        consumed = set()
+        missing = []
+        for index, value in enumerate(v):
+            found = False
+            for i, s in enumerate(self._schemas):
+                if i in consumed:
+                    continue
+                try:
+                    s(value)
+                except Invalid:
+                    pass
+                else:
+                    found = True
+                    consumed.add(i)
+                    break
+            if not found:
+                missing.append((index, value))
+
+        if len(missing) == 1:
+            el = missing[0]
+            raise Invalid(self.msg or 'Element #{} ({}) is not valid against any validator'.format(el[0], el[1]))
+        elif missing:
+            raise MultipleInvalid([
+                Invalid(self.msg or 'Element #{} ({}) is not valid against any validator'.format(el[0], el[1]))
+                for el in missing
+            ])
+        return v
+
+    def __repr__(self):
+        return 'Unordered([{}])'.format(", ".join(repr(v) for v in self.validators))
+
+
+class Number(object):
+    """
+    Verify the number of digits that are present in the number(Precision),
+    and the decimal places(Scale)
+
+    :raises Invalid: If the value does not match the provided Precision and Scale.
+
+    >>> schema = Schema(Number(precision=6, scale=2))
+    >>> schema('1234.01')
+    '1234.01'
+    >>> schema = Schema(Number(precision=6, scale=2, yield_decimal=True))
+    >>> schema('1234.01')
+    Decimal('1234.01')
+    """
+
+    def __init__(self, precision=None, scale=None, msg=None, yield_decimal=False):
+        self.precision = precision
+        self.scale = scale
+        self.msg = msg
+        self.yield_decimal = yield_decimal
+
+    def __call__(self, v):
+        """
+        :param v: is a number enclosed with string
+        :return: Decimal number
+        """
+        precision, scale, decimal_num = self._get_precision_scale(v)
+
+        if self.precision is not None and self.scale is not None and\
+            precision != self.precision and scale != self.scale:
+            raise Invalid(self.msg or "Precision must be equal to %s, and Scale must be equal to %s" %(self.precision, self.scale))
+        else:
+            if self.precision is not None and precision != self.precision:
+                raise Invalid(self.msg or "Precision must be equal to %s"%self.precision)
+
+            if self.scale is not None and scale != self.scale :
+                raise Invalid(self.msg or "Scale must be equal to %s"%self.scale)
+
+        if self.yield_decimal:
+            return decimal_num
+        else:
+            return v
+
+    def __repr__(self):
+        return ('Number(precision=%s, scale=%s, msg=%s)' % (self.precision, self.scale, self.msg))
+
+    def _get_precision_scale(self, number):
+        """
+        :param number:
+        :return: tuple(precision, scale, decimal_number)
+        """
+        try:
+            decimal_num = Decimal(number)
+        except InvalidOperation:
+            raise Invalid(self.msg or 'Value must be a number enclosed with string')
+
+        return (len(decimal_num.as_tuple().digits), -(decimal_num.as_tuple().exponent), decimal_num)
